using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using MeshVR;
using SimpleJSON;
using System.Linq;

namespace McKlinton
{
    public class Viewport
    {
        public Vector3 cameraPosition;
        public Quaternion cameraRotation;
        public Vector3 cameraLocalPosition;
        public Quaternion cameraLocalRotation;
        public Vector3 navRigPosition;
        public Quaternion navRigRotation;
        public Vector3 navRigLocalPosition;
        public Quaternion navRigLocalRotation;
        public Vector3 centerTargetPosition;
        public Vector3 centerTargetLocalPosition;
        public Quaternion centerTargetRotation;
        public Quaternion centerTargetLocalRotation;
        public float focusDistance;
        public float playerHeightAdjust;


        public JSONClass ToJSON()
        {
            JSONClass json = new JSONClass();

            json["cameraPosition"] = Utils.Vector3ToJSON(cameraPosition);
            json["cameraRotation"] = Utils.QuaternionToJSON(cameraRotation);
            json["cameraLocalPosition"] = Utils.Vector3ToJSON(cameraLocalPosition);
            json["cameraLocalRotation"] = Utils.QuaternionToJSON(cameraLocalRotation);
            json["navRigPosition"] = Utils.Vector3ToJSON(navRigPosition);
            json["navRigRotation"] = Utils.QuaternionToJSON(navRigRotation);
            json["navRigLocalPosition"] = Utils.Vector3ToJSON(navRigLocalPosition);
            json["navRigLocalRotation"] = Utils.QuaternionToJSON(navRigLocalRotation);
            json["centerTargetPosition"] = Utils.Vector3ToJSON(centerTargetPosition);
            json["centerTargetLocalPosition"] = Utils.Vector3ToJSON(centerTargetLocalPosition);
            json["centerTargetRotation"] = Utils.QuaternionToJSON(centerTargetRotation);
            json["centerTargetLocalRotation"] = Utils.QuaternionToJSON(centerTargetLocalRotation);
            json["focusDistance"] = new JSONData(focusDistance);
            json["playerHeightAdjust"] = new JSONData(playerHeightAdjust);

            return json;
        }
        private static Vector3 JSONToVector3(JSONClass json)
        {
            return new Vector3(json["x"].AsFloat, json["y"].AsFloat, json["z"].AsFloat);
        }

        private static Quaternion JSONToQuaternion(JSONClass json)
        {
            return new Quaternion(json["x"].AsFloat, json["y"].AsFloat, json["z"].AsFloat, json["w"].AsFloat);
        }

        public static Viewport FromJSON(JSONClass json)
        {
            Viewport viewport = new Viewport
            {
                cameraPosition = JSONToVector3(json["cameraPosition"].AsObject),
                cameraRotation = JSONToQuaternion(json["cameraRotation"].AsObject),
                cameraLocalPosition = JSONToVector3(json["cameraLocalPosition"].AsObject),
                cameraLocalRotation = JSONToQuaternion(json["cameraLocalRotation"].AsObject),
                navRigPosition = JSONToVector3(json["navRigPosition"].AsObject),
                navRigRotation = JSONToQuaternion(json["navRigRotation"].AsObject),
                navRigLocalPosition = JSONToVector3(json["navRigLocalPosition"].AsObject),
                navRigLocalRotation = JSONToQuaternion(json["navRigLocalRotation"].AsObject),
                centerTargetPosition = JSONToVector3(json["centerTargetPosition"].AsObject),
                centerTargetLocalPosition = JSONToVector3(json["centerTargetLocalPosition"].AsObject),
                centerTargetRotation = JSONToQuaternion(json["centerTargetRotation"].AsObject),
                centerTargetLocalRotation = JSONToQuaternion(json["centerTargetLocalRotation"].AsObject),
                focusDistance = json["focusDistance"].AsFloat,
                playerHeightAdjust = json["playerHeightAdjust"].AsFloat
            };

            return viewport;
        }        
    }

    public static class CameraController {

        public static Viewport SnapshotViewport() {
            var sc = SuperController.singleton;
            Viewport viewport = new Viewport();
            viewport.centerTargetPosition = sc.centerCameraTarget.transform.position;
            viewport.centerTargetRotation = sc.centerCameraTarget.transform.rotation;
            viewport.centerTargetLocalPosition = sc.centerCameraTarget.transform.localPosition;
            viewport.centerTargetLocalRotation = sc.centerCameraTarget.transform.localRotation;
            viewport.navRigPosition = sc.navigationRig.transform.position;
            viewport.navRigRotation = sc.navigationRig.transform.rotation;
            viewport.navRigLocalPosition = sc.navigationRig.transform.localPosition;
            viewport.navRigLocalRotation = sc.navigationRig.transform.localRotation;
            viewport.cameraPosition = sc.MonitorCenterCamera.transform.position;
            viewport.cameraRotation = sc.MonitorCenterCamera.transform.rotation;
            viewport.cameraLocalPosition = sc.MonitorCenterCamera.transform.localPosition;
            viewport.cameraLocalRotation = sc.MonitorCenterCamera.transform.localRotation;
            viewport.focusDistance = sc.focusDistance;
            viewport.playerHeightAdjust = sc.playerHeightAdjust;
            return viewport;
        }

        public static void ApplyViewport(Viewport viewport) {
            var sc = SuperController.singleton;
            sc.ResetNavigationRigPositionRotation();

            sc.centerCameraTarget.transform.position = viewport.centerTargetPosition;
            sc.centerCameraTarget.transform.rotation = viewport.centerTargetRotation;
            sc.centerCameraTarget.transform.localPosition = viewport.centerTargetLocalPosition;
            sc.centerCameraTarget.transform.localRotation = viewport.centerTargetLocalRotation;
            sc.navigationRig.transform.position = viewport.navRigPosition;
            sc.navigationRig.transform.rotation = viewport.navRigRotation;
            sc.navigationRig.transform.localPosition = viewport.navRigLocalPosition;
            sc.navigationRig.transform.localRotation = viewport.navRigLocalRotation;
            sc.MonitorCenterCamera.transform.position = viewport.cameraPosition;
            sc.MonitorCenterCamera.transform.rotation = viewport.cameraRotation;
            sc.MonitorCenterCamera.transform.localPosition = viewport.cameraLocalPosition;
            sc.MonitorCenterCamera.transform.localRotation = viewport.cameraLocalRotation;

            sc.focusDistance = viewport.focusDistance;
            sc.playerHeightAdjust = viewport.playerHeightAdjust;
            
            // if(singleton.isOpenVR) cameraTransform = singleton.ViveCenterCamera.transform;
            // else if (singleton.isOVR) cameraTransform = singleton.OVRCenterCamera.transform;
            // else if (singleton.IsMonitorOnly) cameraTransform = singleton.MonitorCenterCamera.transform;
        }

        public static Viewport InterpolateViewport(Viewport start, Viewport end, float t) {
            Viewport viewport = new Viewport();
            viewport.centerTargetPosition = Vector3.Lerp(start.centerTargetPosition, end.centerTargetPosition, t);
            viewport.centerTargetRotation = Quaternion.Slerp(start.centerTargetRotation, end.centerTargetRotation, t);
            viewport.centerTargetLocalPosition = Vector3.Lerp(start.centerTargetLocalPosition, end.centerTargetLocalPosition, t);
            viewport.centerTargetLocalRotation = Quaternion.Slerp(start.centerTargetLocalRotation, end.centerTargetLocalRotation, t);
            viewport.navRigPosition = Vector3.Lerp(start.navRigPosition, end.navRigPosition, t);
            viewport.navRigRotation = Quaternion.Slerp(start.navRigRotation, end.navRigRotation, t);
            viewport.navRigLocalPosition = Vector3.Lerp(start.navRigLocalPosition, end.navRigLocalPosition, t);
            viewport.navRigLocalRotation = Quaternion.Slerp(start.navRigLocalRotation, end.navRigLocalRotation, t);
            viewport.cameraPosition = Vector3.Lerp(start.cameraPosition, end.cameraPosition, t);
            viewport.cameraRotation = Quaternion.Slerp(start.cameraRotation, end.cameraRotation, t);
            viewport.cameraLocalPosition = Vector3.Lerp(start.cameraLocalPosition, end.cameraLocalPosition, t);
            viewport.cameraLocalRotation = Quaternion.Slerp(start.cameraLocalRotation, end.cameraLocalRotation, t);
            viewport.focusDistance = Mathf.Lerp(start.focusDistance, end.focusDistance, t);
            viewport.playerHeightAdjust = Mathf.Lerp(start.playerHeightAdjust, end.playerHeightAdjust, t);
            return viewport;
        }
    }
}