using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using MeshVR;
using SimpleJSON;
using System.Linq;

namespace McKlinton
{
    public class SnapshotSettings
    {
        public bool includePlayerView = true;
        public bool includePose = true;
        public bool includeMorphs = true;
        public bool includeLights = true;
        public bool includeClothing = true; // TODO?

        public static SnapshotSettings fromJSON(JSONClass json)
        {
            SnapshotSettings settings = new SnapshotSettings();
            settings.includePlayerView = json["includePlayerView"].AsBool;
            settings.includePose = json["includePose"].AsBool;
            settings.includeMorphs = json["includeMorphs"].AsBool;
            settings.includeLights = json["includeLights"].AsBool;
            settings.includeClothing = json["includeClothing"].AsBool;
            return settings;
        }
        public JSONClass ToJSON()
        {
            JSONClass json = new JSONClass();
            json["includePlayerView"] = new JSONData(includePlayerView);
            json["includePose"] = new JSONData(includePose);
            json["includeMorphs"] = new JSONData(includeMorphs);
            json["includeLights"] = new JSONData(includeLights);
            json["includeClothing"] = new JSONData(includeClothing);
            return json;
        }
    }

    public class Snapshot
    {
        private MVRScript _parent;
        public string Name = "Snapshot";
        private SnapshotSettings _settings;
        private Dictionary<string, JSONClass> _personSnapshots = new Dictionary<string, JSONClass>();
        private Dictionary<string, JSONClass> _lightSnapshots = new Dictionary<string, JSONClass>();
        private Viewport _viewportSnapshot = new Viewport();

        public Snapshot(MVRScript parent)
        {
            _parent = parent;
            _settings = new SnapshotSettings();
        }

        public Snapshot(MVRScript parent, JSONClass json, string name)
        {
            _parent = parent;
            foreach (KeyValuePair<string, JSONNode> person in json["people"].AsObject)
            {
                _personSnapshots[person.Key] = person.Value.AsObject;
            }
            _viewportSnapshot = Viewport.FromJSON(json["viewport"].AsObject);
            _settings = SnapshotSettings.fromJSON(json["snapshotSettings"].AsObject);
        }

        public void RestoreFromJSON(JSONClass json, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            foreach (KeyValuePair<string, JSONNode> person in json["people"].AsObject)
            {
                _personSnapshots[person.Key] = person.Value.AsObject;
            }
            foreach (KeyValuePair<string, JSONNode> light in json["lights"].AsObject)
            {
                _lightSnapshots[light.Key] = light.Value.AsObject;
            }
            Name = json["name"].Value;
            _viewportSnapshot = Viewport.FromJSON(json["viewport"].AsObject);
            _settings = SnapshotSettings.fromJSON(json["snapshotSettings"].AsObject);
        }
        
        public JSONClass ToJSON()
        {
            JSONClass json = new JSONClass();

            JSONClass people = new JSONClass();
            foreach (KeyValuePair<string, JSONClass> snapshot in _personSnapshots)
            {
                people[snapshot.Key] = snapshot.Value;
            }
            JSONClass lights = new JSONClass();
            foreach (KeyValuePair<string, JSONClass> snapshot in _lightSnapshots)
            {
                lights[snapshot.Key] = snapshot.Value;
            }

            json["name"] = Name;
            json["people"] = people;
            json["lights"] = lights;
            json["viewport"] = _viewportSnapshot.ToJSON();
            json["snapshotSettings"] = _settings.ToJSON();
            return json;
        }

        public SnapshotSettings GetSettings()
        {
            return _settings;
        }

        public Dictionary<string, JSONClass> GetPersonSnapshots()
        {
            return _personSnapshots;
        }

        public Viewport GetViewport()
        {
            return _viewportSnapshot;
        }

        public void PrintDebugInfo()
        {
            // Print names of all snapshots
            foreach (var snapshot in _personSnapshots)
            {
                Log.Info("Saved data:" + snapshot.Key);
            }

            // Print settings
            Log.Info("Settings: " + _settings.ToJSON().ToString());

        }
        public void TakeSnapshot(SnapshotSettings settings)
        {
            _settings = settings;

            if (_settings.includePose || _settings.includeMorphs) {
                var persons = SuperController.singleton.GetAtoms().Where(atom => atom.category == "People");

                foreach (var person in persons)
                {
                    JSONClass poseJSON = FromAtom(person, true, _settings.includePose, _settings.includeMorphs, false);
                    _personSnapshots[person.name] = poseJSON;
                }
            }

            if (_settings.includePlayerView)
            {
                var sc = SuperController.singleton;

                var camera = sc.MonitorCenterCamera.transform;
                var navigationRig = sc.navigationRig.transform;

                _viewportSnapshot = CameraController.SnapshotViewport();
            }

            if (_settings.includeLights)
            {
                var lights = SuperController.singleton.GetAtoms().Where(atom => atom.category == "Light");
                Log.Info("Lights: " + lights.Count());
                foreach (var light in lights)
                {
     
                    var lightStorable = light.GetStorableByID("Light") as JSONStorable;
                    var controlStorable = light.GetStorableByID("control") as JSONStorable;

                    JSONClass storables = new JSONClass();
                    storables["Light"] = lightStorable.GetJSON(true, true, true);
                    storables["control"] = controlStorable.GetJSON(true, false, false);
                    _lightSnapshots[light.name] = storables;
                }

            }
        }


        public void ApplySnapshot()
        {
            if (_settings.includePose || _settings.includeMorphs)
            {
                var persons = SuperController.singleton.GetAtoms().Where(atom => atom.category == "People");

                foreach (var person in persons)
                {
                    if (_personSnapshots.ContainsKey(person.name))
                    {
                        JSONClass poseJSON = _personSnapshots[person.name];
                        LoadSnapshot(person, poseJSON, false);
                    }
                }
            }

            if (_settings.includePlayerView)
            {
                CameraController.ApplyViewport(_viewportSnapshot);
            }

            if (_settings.includeLights)
            {
                var lights = SuperController.singleton.GetAtoms().Where(atom => atom.category == "Light");
                foreach (var light in lights)
                {
                    if (_lightSnapshots.ContainsKey(light.name))
                    {
                        JSONClass lightJSON = _lightSnapshots[light.name];
                        var lightMainController = light.GetStorableByID("control") as JSONStorable;
                        lightMainController.transform.position = Utils.JSONToVector3(lightJSON["control"]["position"].AsObject);
                        lightMainController.transform.rotation = Quaternion.Euler(Utils.JSONToVector3(lightJSON["control"]["rotation"].AsObject));

                        // Other light settings
                        var lightStorable = light.GetStorableByID("Light") as JSONStorable;
                        lightStorable.RestoreFromJSON(lightJSON["Light"].AsObject);

                    }
                }
            }
        }
        
         // Following methods for saving and loading pose presets are mostly borrowed from AcidBubbles Timeline/AtomPose.cs. 
         // Thanks to AcidBubbles figuring out how to do this.
        public JSONClass FromAtom(Atom atom, bool includeRoot, bool includePose, bool includeMorphs, bool useMergeLoad)
        {
            if (atom.type != "Person") return null;

            var storables = atom.GetStorableIDs()
                .Select(atom.GetStorableByID)
                .Where(t => !t.exclude && t.gameObject.activeInHierarchy)
                .Where(t => ShouldStorableBeIncluded(t, includeRoot, includePose, includeMorphs));
            var storablesJSON = new JSONArray();
            foreach (var storable in storables)
            {
                var storableJSON = storable.GetJSON(true, false, storable.name != "geometry");
                storablesJSON.Add(storableJSON);
            }

            var poseJSON = new JSONClass
            {
                ["clipPoseSettings"] = new JSONClass
                {
                    ["includeRoot"] = { AsBool = includeRoot },
                    ["includePose"] = { AsBool = includePose },
                    ["includeMorphs"] = { AsBool = includeMorphs },
                    ["useMergeLoad"] = { AsBool = useMergeLoad },
                },
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["V2"] = { AsBool = true },
                ["storables"] = storablesJSON
            };

            return poseJSON;
        }
    
        private bool ShouldStorableBeIncluded(JSONStorable t, bool includeRoot, bool includePose, bool includeMorphs)
        {
            if (includePose && t is FreeControllerV3) return includeRoot || ((FreeControllerV3) t).name != "control";
            if (includePose && t is DAZBone) return true;
            if (includeMorphs && t.storeId == "geometry" && t is DAZCharacterSelector)return true;

            return false;
        }

        private void LoadSnapshot(Atom atom, JSONClass poseJSON, bool useMergeLoad)
        {
            var posePresetsManagerControls = atom.presetManagerControls.First(pmc => pmc.name == "PosePresets");
            var posePresetsManager = posePresetsManagerControls.GetComponent<PresetManager>();
            posePresetsManager.LoadPresetFromJSON(poseJSON, useMergeLoad);
        }
    }
}