using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.SceneManagement;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;

namespace McKlinton
{
    public class GlobalSettings
    {
        public SnapshotSettings snapshotSettings;
        public float UIScaleFactor = 1.0f;
        public string ShowDesktopUIShortcut = "None";
        public string SaveSnapshotShortcut = "None";
        public string OverwriteSnapshotShortcut = "None";
        public string DeleteSnapshotShortcut = "None";


        public static GlobalSettings fromJSON(JSONClass json)
        {
            var settings = new GlobalSettings();
            settings.snapshotSettings = SnapshotSettings.fromJSON(json["snapshotSettings"].AsObject);
            settings.UIScaleFactor = json["UIScaleFactor"].AsFloat;
            settings.ShowDesktopUIShortcut = json["shortcuts"]["ShowDesktopUIShortcut"];
            settings.SaveSnapshotShortcut = json["shortcuts"]["SaveSnapshotShortcut"];
            settings.OverwriteSnapshotShortcut = json["shortcuts"]["OverwriteSnapshotShortcut"];
            settings.DeleteSnapshotShortcut = json["shortcuts"]["DeleteSnapshotShortcut"];
            return settings;
        }
        public JSONClass ToJSON()
        {
            var json = new JSONClass();
            json["snapshotSettings"] = snapshotSettings.ToJSON();
            json["UIScaleFactor"] = UIScaleFactor.ToString("r");
            json["shortcuts"] = new JSONClass();
            json["shortcuts"]["ShowDesktopUIShortcut"] = ShowDesktopUIShortcut;
            json["shortcuts"]["SaveSnapshotShortcut"] = SaveSnapshotShortcut;
            json["shortcuts"]["OverwriteSnapshotShortcut"] = OverwriteSnapshotShortcut;
            json["shortcuts"]["DeleteSnapshotShortcut"] = DeleteSnapshotShortcut;

            return json;
        }
    }

    public static class Log
    {
        private const bool LOGGING_ENABLED = false;
        public static void Info(string message)
        {
            if (LOGGING_ENABLED)
            {
                SuperController.LogMessage($"SnapshotManager: {message}");
            }
        }
        public static void Error(string message)
        {
            SuperController.LogError($"SnapshotManager: {message}");
        }
    }

    public interface ISnapshotManager
    {
        List<int> GetSnapshotIds();
        Snapshot GetSnapshot(int id);
        void SaveSnapshotName(int id, string name);
    }

    public class SnapshotManager : MVRScript, ISnapshotManager
    {
        private bool _enabled = true;
        private bool _started = false;
        private GlobalSettings _globalSettings;
        private SnapshotManagerUI _snapshotManagerUI;
        private Dictionary<int, Snapshot> _snapshots = new Dictionary<int, Snapshot>();
        private int _snapshotIndex = 1;
        private int _currentSnapshotId = -1;

        // VAM UI elements
        private JSONStorableBool _showDesktopUiJSON;
        // UI Size storables
        private JSONStorableFloat _uiScaleJSON;
        private JSONStorableStringChooser _showDesktopUiShortCutJSON;
        private JSONStorableStringChooser _saveSnapshotShortCutJSON;
        private JSONStorableStringChooser _overwriteSnapshotShortCutJSON;
        private JSONStorableStringChooser _deleteSnapshotShortCutJSON;
        
        // SnapshotManager trigger storables
        private JSONStorableString _activeSnapshotByNameJSON;
        private List<JSONStorableAction> _triggerStorables = new List<JSONStorableAction>();

        // SnapshotManager interface
        public List<int> GetSnapshotIds()
        {
            return _snapshots.Keys.ToList();
        }
        public Snapshot GetSnapshot(int id)
        {
            Log.Info($"Getting snapshot: {id}");
            if (!_snapshots.ContainsKey(id))
            {
                Log.Error($"Snapshot not found: {id}");
                return null;
            }
            return _snapshots[id];
        }
        public void SaveSnapshotName(int id, string name)
        {
            Log.Info($"Saving snapshot name: {id} - {name}");
            if (!_snapshots.ContainsKey(id))
            {
                Log.Error($"Snapshot not found: {id}");
                return;
            }
            _snapshots[id].Name = name;
            _snapshotManagerUI.UpdateUI(_snapshots);
        }
        public override void Init()
        {
            pluginLabelJSON.val = "SnapshotManager - McKlinton";

            // Check if plugin is already running as another instance
            var existingUI = GameObject.Find("McKlintonSnapshotManagerUI");
            if (existingUI != null)
            {
                Log.Error($"SnapshotManager plugin is already running");
                return;
            }
            else {
                _globalSettings = new GlobalSettings();
                _globalSettings = RestoreSettingsFromFile();
                if (_globalSettings.snapshotSettings == null)
                {
                    _globalSettings.snapshotSettings = new SnapshotSettings();
                }
                _snapshots = new Dictionary<int, Snapshot>();
                var pluginDataPath = Utils.GetPackagePluginDataPath(this);
                _snapshotManagerUI = new SnapshotManagerUI( _globalSettings, this, saveSnapshot, applySnapshot, overwriteSnapshot, deleteSnapshot, pluginDataPath, _globalSettings.UIScaleFactor);
                SetupTriggerStorables();
                SetupUI();
                _started = true;
            }
        }

        void Start()
        {
            if (_started)
            {
                if (_snapshots.Count > 0)
                {
                    _snapshotIndex = _snapshots.Keys.Max() + 1;
                }
                Log.Info($"Loaded {_snapshots.Count} snapshots");
                _snapshotManagerUI.UpdateUI(_snapshots);
                _showDesktopUiJSON.val = true;
            }
        }
        private JSONClass snapshotsToJSON()
        {
            var json = new JSONClass();
            foreach (var snapshot in _snapshots)
            {
                json[snapshot.Key.ToString()] = snapshot.Value.ToJSON();
            }
            return json;
        }
        
        void OnDestroy()
        {
            try
            {
                if (_started)
                {
                    _snapshotManagerUI.OnDestroy();
                    _snapshotManagerUI = null;
                }
            }
            catch (Exception e)
            {
                Log.Error($"Error: {e}");

                // Make sure the UI is destroyed
                var existingUI = GameObject.Find("McKlintonSnapshotManagerUI");
                if (existingUI != null)
                {
                    GameObject.Destroy(existingUI);
                    existingUI = null;
                }
            }
        }

        public void Update()
        {
            if (_started){

                if (Input.GetKeyDown(Utils.GetKeyCode(_globalSettings.ShowDesktopUIShortcut)))
                {
                    _showDesktopUiJSON.val = !_showDesktopUiJSON.val;
                }
                if (Input.GetKeyDown(Utils.GetKeyCode(_globalSettings.SaveSnapshotShortcut)))
                {
                    saveSnapshot();
                }
                if (Input.GetKeyDown(Utils.GetKeyCode(_globalSettings.OverwriteSnapshotShortcut)))
                {
                    if (_currentSnapshotId != -1)
                    {
                        overwriteSnapshot(_currentSnapshotId);
                    }
                }
                if (Input.GetKeyDown(Utils.GetKeyCode(_globalSettings.DeleteSnapshotShortcut)))
                {
                    if (_currentSnapshotId != -1)
                    {
                        deleteSnapshot(_currentSnapshotId);
                    }
                }
            }
        }

        private void saveSnapshot(int id=-1)
        {
            var snapshot = new Snapshot(this);
            var settings = new SnapshotSettings();
            settings.includePlayerView = _globalSettings.snapshotSettings.includePlayerView;
            settings.includePose = _globalSettings.snapshotSettings.includePose;
            settings.includeMorphs = _globalSettings.snapshotSettings.includeMorphs;
            settings.includeLights = _globalSettings.snapshotSettings.includeLights;
            settings.includeClothing = _globalSettings.snapshotSettings.includeClothing;

            snapshot.TakeSnapshot(settings);
            snapshot.Name = "snapshot"+_snapshotIndex;

            if (id == -1)
            {
                id = _snapshotIndex;
                _snapshotIndex++;
            }
            _snapshots[id] = snapshot;
            _snapshotManagerUI.UpdateUI(_snapshots);

            // Create trigger storable
            var triggerStorable = new JSONStorableAction("ActivateSnapshot_ID"+id, () =>
            {
                applySnapshot(id);
            });
            RegisterAction(triggerStorable);
            _triggerStorables.Add(triggerStorable);
        }

        private void applySnapshot(int id)
        {
            Log.Info($"Applying snapshot: {id}");
            var snapshot = _snapshots[id];
            snapshot.ApplySnapshot();
            _currentSnapshotId = id;
        }

        private void deleteSnapshot(int id)
        {
            Log.Info($"Deleting snapshot: {id}");
            _snapshots.Remove(id);
            _snapshotManagerUI.UpdateUI(_snapshots);

            // Remove trigger storable
            var triggerStorable = _triggerStorables.FirstOrDefault(x => x.name == "ActivateSnapshot_ID"+id);
            if (triggerStorable != null)
            {
                Log.Info($"Removing trigger storable: {triggerStorable.name}");
                DeregisterAction(triggerStorable);
            }
        }

        private void overwriteSnapshot(int id)
        {
            Log.Info($"Overwriting snapshot: {id}");
            var snapshot = _snapshots[id];

            var settings = new SnapshotSettings();
            settings.includePlayerView = _globalSettings.snapshotSettings.includePlayerView;
            settings.includePose = _globalSettings.snapshotSettings.includePose;
            settings.includeMorphs = _globalSettings.snapshotSettings.includeMorphs;
            settings.includeLights = _globalSettings.snapshotSettings.includeLights;
            settings.includeClothing = _globalSettings.snapshotSettings.includeClothing;

            snapshot.TakeSnapshot(settings);
            _snapshotManagerUI.UpdateUI(_snapshots);
        }

        private void OnActiveChange(bool enabled)
        {
            _enabled = enabled;
            _snapshotManagerUI.SetVisible(enabled);
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            try
            {
                json["Snapshots"] = snapshotsToJSON();

            }
            catch (Exception exc)
            {
                Log.Error($"Exception caught: {exc}");
            }

            return json;
        }
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!jc.HasKey("Snapshots"))
            {
                Log.Info($"No snapshots found");
                return;
            }

            var snapshotsJSON = jc["Snapshots"];
            if (snapshotsJSON != null && snapshotsJSON.AsObject != null)
            {
                foreach (KeyValuePair<string, JSONNode> item in snapshotsJSON.AsObject)
                {
                    string key = item.Key;
                    JSONClass value = item.Value.AsObject;

                    var snapshot = new Snapshot(this);
                    snapshot.RestoreFromJSON(value, base.subScenePrefix, base.mergeRestore, setMissingToDefault);

                    int snapshotKey = int.Parse(key);
                    _snapshots[snapshotKey] = snapshot;
                }
            }
        }

        private void SaveSettingsTofile()
        {
            var json = _globalSettings.ToJSON();

            if (!FileManagerSecure.DirectoryExists(Utils.PluginDataDirectoryLocalPath))
            {
                FileManagerSecure.CreateDirectory(Utils.PluginDataDirectoryLocalPath);
            }
            var path = Utils.PluginDataDirectoryLocalPath + "default_settings.json";
            FileManagerSecure.WriteAllText(path, json.ToString());
        }

        private GlobalSettings RestoreSettingsFromFile()
        {
            var path = Utils.PluginDataDirectoryLocalPath + "default_settings.json";
            if (FileManagerSecure.FileExists(path))
            {
                var json = FileManagerSecure.ReadAllText(path);
                return GlobalSettings.fromJSON(JSON.Parse(json).AsObject);
            }
            return new GlobalSettings();
        }

        private void SetupKeyCodeChooser(JSONStorableStringChooser json, string label, string key, Action<string> callback, string defaultKey=null)
        {
            var keyCodes = Utils.GetKeyCodes();
            var keyCodesList = new List<string>(keyCodes);
            json = new JSONStorableStringChooser(label, keyCodesList, keyCodesList[0], label)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            CreateScrollablePopup(json, false);
            RegisterStringChooser(json);
            json.setCallbackFunction += (string val) =>
            {
                callback(val);
            };

            if (defaultKey != null)
            {
                json.val = defaultKey;
            }
        }
        private void SetupUI()
        {
            _showDesktopUiJSON = new JSONStorableBool("Show UI", false)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            CreateToggle(_showDesktopUiJSON, false);
            RegisterBool(_showDesktopUiJSON);

            _showDesktopUiJSON.setCallbackFunction += OnActiveChange;


            // UI Scale slider
            _uiScaleJSON = new JSONStorableFloat("UI Scale Factor", _globalSettings.UIScaleFactor, 0.1f, 2.0f, false, true)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            CreateSlider(_uiScaleJSON, false);
            RegisterFloat(_uiScaleJSON);

            // callback for UI scale change
            _uiScaleJSON.setCallbackFunction += (float val) =>
            {
                _globalSettings.UIScaleFactor = val;
                _snapshotManagerUI.SetScale(val);
                _snapshotManagerUI.UpdateUI(_snapshots);
            };
            
            // ShowDesktopUIShortcut
            SetupKeyCodeChooser(
                _showDesktopUiShortCutJSON, 
                "Show Desktop UI Shortcut", 
                _globalSettings.ShowDesktopUIShortcut,
                (string val) => {
                    _globalSettings.ShowDesktopUIShortcut = val;
                }, 
                _globalSettings.ShowDesktopUIShortcut);

            // SaveSnapshotShortcut
            SetupKeyCodeChooser(
                _saveSnapshotShortCutJSON, 
                "Save Snapshot Shortcut", 
                _globalSettings.SaveSnapshotShortcut,
                (string val) => {
                    _globalSettings.SaveSnapshotShortcut = val;
                }, 
                _globalSettings.SaveSnapshotShortcut);
            
            // OverwriteSnapshotShortcut
            SetupKeyCodeChooser(
                _overwriteSnapshotShortCutJSON, 
                "Overwrite Snapshot Shortcut", 
                _globalSettings.OverwriteSnapshotShortcut,
                (string val) => {
                    _globalSettings.OverwriteSnapshotShortcut = val;
                }, 
                _globalSettings.OverwriteSnapshotShortcut);

            // DeleteSnapshotShortcut
            SetupKeyCodeChooser(
                _deleteSnapshotShortCutJSON, 
                "Delete Snapshot Shortcut", 
                _globalSettings.DeleteSnapshotShortcut,
                (string val) => {
                    _globalSettings.DeleteSnapshotShortcut = val;
                }, 
                _globalSettings.DeleteSnapshotShortcut);

            // Button to save settings
            var saveSettingsButton = CreateButton("Save as default");
            saveSettingsButton.button.onClick.AddListener(() =>
            {
               SaveSettingsTofile();
            });
        }

        private void SetupTriggerStorables()
        {
            // TODO: Not yet working perfectly, triggers only once
            // _activeSnapshotByNameJSON = new JSONStorableString("ActivateSnapshotByName", "");
            // RegisterString(_activeSnapshotByNameJSON);

            // _activeSnapshotByNameJSON.setCallbackFunction += (string val) =>
            // {
            //     Log.Info($"Activating snapshot by name: {val}");
            //     var snapshot = _snapshots.FirstOrDefault(x => x.Value.Name == val);
            //     if (snapshot.Value != null)
            //     {
            //         applySnapshot(snapshot.Key);
            //     }
            // };
        }

    }
}
