using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq; 


namespace McKlinton
{

    public class SnapshotManagerUI : MonoBehaviour
    {
        private GlobalSettings _globalSettings;
        private float _scale = 1.0f;
        private ISnapshotManager _snapshotManager;
        private GameObject _rootGameObject = null;
        private GameObject _panelShadowGameObject;
        private GameObject _panelGameObject;
        private GameObject _settingsPanelGameObject;
        private bool _settingsPanelCreated = false;
        private GameObject _snapshotSettingsPanelGameObject;
        private Canvas _mainCanvas;
        private string _cameraIconFile = "camera_icon.png";
        private string _poseIconFile = "pose_icon.png";
        private string _morphIconFile = "morph_icon.png";
        private string _toggleOnFile = "checked.png";
        private string _lightBulbIconFile = "light_icon.png";
        private string _settingsIconFile = "settings.png";
        private Texture2D _cameraIcon;
        private Texture2D _poseIcon;
        private Texture2D _morphIcon;
        private Texture2D _toggleOn;
        private Texture2D _lightBulbIcon;
        private Texture2D _settingsIcon;

        private bool _visible = false;

        private bool _settingsPanelActive = false;
        private int _snapshotIdActiveInSettingsPanel = -1;
        private InputField _snapshotNameInputField;

       // Actions
        private Action<int> _onSaveSnapshotClicked;
        private Action<int> _onApplyClicked;
        private Action<int> _onOverwriteClicked;
        private Action<int> _onDeleteClicked;

        public SnapshotManagerUI(
            GlobalSettings globalSettings,
            ISnapshotManager snapshotManager,
            Action<int> onSnapshotButtonClicked, 
            Action<int> onApplyButtonClicked, 
            Action<int> onOverwriteButtonClicked, 
            Action<int> onDeleteButtonClicked,
            string pluginDataPath,
            float scale = 1.0f
            )
        {

            Log.Info("SnapshotManagerUI constructor, using pluginDataPath: " + pluginDataPath);
            _globalSettings = globalSettings;
            _snapshotManager = snapshotManager;
            _scale = scale;
            _onSaveSnapshotClicked = onSnapshotButtonClicked;
            _onApplyClicked = onApplyButtonClicked;
            _onOverwriteClicked = onOverwriteButtonClicked;
            _onDeleteClicked = onDeleteButtonClicked;
            
            _cameraIcon = UIHelper.loadTextureFromFile(pluginDataPath + _cameraIconFile);
            _poseIcon = UIHelper.loadTextureFromFile(pluginDataPath + _poseIconFile);
            _morphIcon = UIHelper.loadTextureFromFile(pluginDataPath + _morphIconFile);
            _toggleOn = UIHelper.loadTextureFromFile(pluginDataPath + _toggleOnFile);
            _lightBulbIcon = UIHelper.loadTextureFromFile(pluginDataPath + _lightBulbIconFile);
            _settingsIcon = UIHelper.loadTextureFromFile(pluginDataPath + _settingsIconFile);

            EnsureEventSystemExists();

            // Create the Canvas
            _rootGameObject = new GameObject("McKlintonSnapshotManagerUI");
            _rootGameObject.SetActive(false);
            _mainCanvas = _rootGameObject.AddComponent<Canvas>();
            _mainCanvas.sortingLayerName = "BeforeDefault";
            _mainCanvas.sortingOrder = 0;
            _mainCanvas.renderMode = RenderMode.ScreenSpaceCamera;
            _mainCanvas.worldCamera = Camera.main;

            CanvasScaler canvasScaler = _rootGameObject.AddComponent<CanvasScaler>();
            canvasScaler.uiScaleMode = CanvasScaler.ScaleMode.ConstantPixelSize;
            canvasScaler.scaleFactor = _scale;

            _rootGameObject.AddComponent<GraphicRaycaster>();

            SetupUI();
            _rootGameObject.SetActive(true);
        }

        public void SetVisible(bool visible)
        {
            _visible = visible;
            if (_rootGameObject != null)
            {
                Log.Info("Setting SnapshotManagerUI visible: " + _visible);
                _rootGameObject.SetActive(_visible);
            }
        }
        private void EnsureEventSystemExists() {
            if (FindObjectOfType<EventSystem>() == null) {
                GameObject eventSystem = new GameObject("EventSystem");
                eventSystem.AddComponent<EventSystem>();
                eventSystem.AddComponent<StandaloneInputModule>();
            }
        }

        public void SetScale(float scale)
        {
            _scale = scale;
            // Update CanvasScaler
            CanvasScaler canvasScaler = _rootGameObject.GetComponent<CanvasScaler>();
            canvasScaler.scaleFactor = _scale;
        }
        
        private void SetupUI() {
            Log.Info("SnapshotManagerUI start");

            if (_panelShadowGameObject != null) {
                GameObject.Destroy(_panelShadowGameObject);
            }

            // Create a Panel shadow
            _panelShadowGameObject = new GameObject("PanelShadow");
            _panelShadowGameObject.transform.SetParent(_rootGameObject.transform, false);
            _panelShadowGameObject.AddComponent<CanvasRenderer>();
            Image panelShadowImage = _panelShadowGameObject.AddComponent<Image>();
            panelShadowImage.color = SolarizedColors.DarkTheme.Background;
            RectTransform panelShadowRect = _panelShadowGameObject.GetComponent<RectTransform>();
            panelShadowRect.sizeDelta = new Vector2(820, 820);
            panelShadowRect.anchorMin = new Vector2(0, 1);
            panelShadowRect.anchorMax = new Vector2(0, 1);
            panelShadowRect.pivot = new Vector2(0, 1);
            panelShadowRect.anchoredPosition = new Vector2(10, -10);

            // Create a Panel
            _panelGameObject = new GameObject("Panel");
            _panelGameObject.transform.SetParent(_panelShadowGameObject.transform, false);
            _panelGameObject.AddComponent<CanvasRenderer>();
            Image panelImage = _panelGameObject.AddComponent<Image>();
            panelImage.color = SolarizedColors.LightTheme.Foreground;
            RectTransform panelRect = _panelGameObject.GetComponent<RectTransform>();
            panelRect.sizeDelta = new Vector2(800, 800);
            panelRect.anchorMin = new Vector2(0, 1);
            panelRect.anchorMax = new Vector2(0, 1);
            panelRect.pivot = new Vector2(0, 1);
            panelRect.anchoredPosition = new Vector2(10, -10);

            // Create title text
            GameObject titleGameObject = new GameObject("Title");
            titleGameObject.transform.SetParent(_panelGameObject.transform, false);
            Text titleText = titleGameObject.AddComponent<Text>();
            titleText.text = "SNAPSHOT MANAGER";
            titleText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            titleText.color = Color.black;
            titleText.fontSize = 30;
            titleText.fontStyle = FontStyle.BoldAndItalic;
            titleText.alignment = TextAnchor.UpperLeft;
            RectTransform titleRect = titleText.GetComponent<RectTransform>();
            titleRect.sizeDelta = new Vector2(350, 50);
            titleRect.anchorMin = new Vector2(0, 1);
            titleRect.anchorMax = new Vector2(0, 1);
            titleRect.pivot = new Vector2(0, 1);
            titleRect.anchoredPosition = new Vector2(10, -10);

            // Settings button
            GameObject settingsButtonGameObject = CreateButton(new Vector2(500, -10), "Settings", -1, SolarizedColors.Cyan, _settingsIcon, OnSettingsButtonClicked);
            settingsButtonGameObject.transform.SetParent(_panelGameObject.transform, false);

            // Takesnaphot button
            GameObject takeSnapshotButtonGameObject = CreateButton(new Vector2(650, -10), "Snapshot", -1, SolarizedColors.Blue, _cameraIcon, _onSaveSnapshotClicked);
            takeSnapshotButtonGameObject.transform.SetParent(_panelGameObject.transform, false);
        }

        GameObject CreateSettingsPanel(Vector2 pos)
        {
            // Create shadow panel
            GameObject settingsPanelShadowGameObject = new GameObject("SettingsPanelShadow");
            settingsPanelShadowGameObject.AddComponent<CanvasRenderer>();
            Image settingsPanelShadowImage = settingsPanelShadowGameObject.AddComponent<Image>();
            settingsPanelShadowImage.color = SolarizedColors.DarkTheme.Background;
            RectTransform settingsPanelShadowRect = settingsPanelShadowGameObject.GetComponent<RectTransform>();
            settingsPanelShadowRect.sizeDelta = new Vector2(400, 400);
            settingsPanelShadowRect.anchorMin = new Vector2(0, 1);
            settingsPanelShadowRect.anchorMax = new Vector2(0, 1);
            settingsPanelShadowRect.pivot = new Vector2(0, 1);
            settingsPanelShadowRect.anchoredPosition = pos;

            // Create a settings Panel
            GameObject settingsPanelGameObject = new GameObject("SettingsPanel");
            settingsPanelGameObject.transform.SetParent(settingsPanelShadowGameObject.transform, false);
            settingsPanelGameObject.AddComponent<CanvasRenderer>();
            Image settingsPanelImage = settingsPanelGameObject.AddComponent<Image>();
            settingsPanelImage.color = SolarizedColors.LightTheme.Foreground;
            RectTransform settingsPanelRect = settingsPanelGameObject.GetComponent<RectTransform>();
            settingsPanelRect.sizeDelta = new Vector2(380, 380);
            settingsPanelRect.anchorMin = new Vector2(0, 1);
            settingsPanelRect.anchorMax = new Vector2(0, 1);
            settingsPanelRect.pivot = new Vector2(0, 1);
            settingsPanelRect.anchoredPosition = new Vector2(10, -10);

            // Setting title
            GameObject settingsTitleGameObject = new GameObject("SettingsTitle");
            settingsTitleGameObject.transform.SetParent(settingsPanelGameObject.transform, false);
            Text settingsTitleText = settingsTitleGameObject.AddComponent<Text>();
            settingsTitleText.text = "SETTINGS";
            settingsTitleText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            settingsTitleText.color = Color.black;
            settingsTitleText.fontSize = 24;
            settingsTitleText.fontStyle = FontStyle.BoldAndItalic;
            settingsTitleText.alignment = TextAnchor.UpperLeft;
            RectTransform settingsTitleRect = settingsTitleText.GetComponent<RectTransform>();
            settingsTitleRect.sizeDelta = new Vector2(350, 50);
            settingsTitleRect.anchorMin = new Vector2(0, 1);
            settingsTitleRect.anchorMax = new Vector2(0, 1);
            settingsTitleRect.pivot = new Vector2(0, 1);
            settingsTitleRect.anchoredPosition = new Vector2(10, -10);
            
            // Settings
            int toggleStartY = -50;
            int toggleMargin = 35;
            var settings = _globalSettings.snapshotSettings;
            GameObject includeViewportToggleGameObject = CreateToggle("Include Viewport", "includeViewport", settings.includePlayerView, new Vector2(toggleMargin, toggleStartY));
            includeViewportToggleGameObject.transform.SetParent(settingsPanelGameObject.transform, false);
            
            GameObject includePoseToggleGameObject = CreateToggle("Include Pose", "includePose", settings.includePose, new Vector2(toggleMargin, toggleStartY - 35));
            includePoseToggleGameObject.transform.SetParent(settingsPanelGameObject.transform, false);

            GameObject includeMorphsToggleGameObject = CreateToggle("Include Pose Morphs", "includeMorphs", settings.includeMorphs, new Vector2(toggleMargin, toggleStartY - 70));
            includeMorphsToggleGameObject.transform.SetParent(settingsPanelGameObject.transform, false);

            GameObject includeLightsToggleGameObject = CreateToggle("Include Lights", "includeLights", settings.includeLights, new Vector2(toggleMargin, toggleStartY - 105));
            includeLightsToggleGameObject.transform.SetParent(settingsPanelGameObject.transform, false);

            return settingsPanelShadowGameObject;
        }

        private GameObject CreateSnapshotSettingsPanel(Vector2 pos)
        {
            // Create shadow panel
            GameObject snapshotSettingsPanelShadowGameObject = new GameObject("SnapshotSettingsPanelShadow");
            snapshotSettingsPanelShadowGameObject.AddComponent<CanvasRenderer>();
            Image snapshotSettingsPanelShadowImage = snapshotSettingsPanelShadowGameObject.AddComponent<Image>();
            snapshotSettingsPanelShadowImage.color = SolarizedColors.DarkTheme.Background;
            RectTransform snapshotSettingsPanelShadowRect = snapshotSettingsPanelShadowGameObject.GetComponent<RectTransform>();
            snapshotSettingsPanelShadowRect.sizeDelta = new Vector2(400, 400);
            snapshotSettingsPanelShadowRect.anchorMin = new Vector2(0, 1);
            snapshotSettingsPanelShadowRect.anchorMax = new Vector2(0, 1);
            snapshotSettingsPanelShadowRect.pivot = new Vector2(0, 1);
            snapshotSettingsPanelShadowRect.anchoredPosition = pos;

            // Create a snapshot settings Panel
            GameObject snapshotSettingsPanelGameObject = new GameObject("SnapshotSettingsPanel");
            snapshotSettingsPanelGameObject.transform.SetParent(snapshotSettingsPanelShadowGameObject.transform, false);
            snapshotSettingsPanelGameObject.AddComponent<CanvasRenderer>();
            Image snapshotSettingsPanelImage = snapshotSettingsPanelGameObject.AddComponent<Image>();
            snapshotSettingsPanelImage.color = SolarizedColors.LightTheme.Foreground;
            RectTransform snapshotSettingsPanelRect = snapshotSettingsPanelGameObject.GetComponent<RectTransform>();
            snapshotSettingsPanelRect.sizeDelta = new Vector2(380, 380);
            snapshotSettingsPanelRect.anchorMin = new Vector2(0, 1);
            snapshotSettingsPanelRect.anchorMax = new Vector2(0, 1);
            snapshotSettingsPanelRect.pivot = new Vector2(0, 1);
            snapshotSettingsPanelRect.anchoredPosition = new Vector2(10, -10);

            // Setting title
            GameObject snapshotSettingsTitleGameObject = new GameObject("SnapshotSettingsTitle");
            snapshotSettingsTitleGameObject.transform.SetParent(snapshotSettingsPanelGameObject.transform, false);
            Text snapshotSettingsTitleText = snapshotSettingsTitleGameObject.AddComponent<Text>();
            snapshotSettingsTitleText.text = "SNAPSHOT SETTINGS";
            snapshotSettingsTitleText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            snapshotSettingsTitleText.color = Color.black;
            snapshotSettingsTitleText.fontSize = 24;
            snapshotSettingsTitleText.fontStyle = FontStyle.BoldAndItalic;
            snapshotSettingsTitleText.alignment = TextAnchor.UpperLeft;
            RectTransform snapshotSettingsTitleRect = snapshotSettingsTitleText.GetComponent<RectTransform>();
            snapshotSettingsTitleRect.sizeDelta = new Vector2(350, 50);
            snapshotSettingsTitleRect.anchorMin = new Vector2(0, 1);
            snapshotSettingsTitleRect.anchorMax = new Vector2(0, 1);
            snapshotSettingsTitleRect.pivot = new Vector2(0, 1);
            snapshotSettingsTitleRect.anchoredPosition = new Vector2(10, -10);

            // Snaphot ID label
            GameObject snapshotIdLabelGameObject = new GameObject("SnapshotIdLabel");
            snapshotIdLabelGameObject.transform.SetParent(snapshotSettingsPanelGameObject.transform, false);
            Text snapshotIdLabelText = snapshotIdLabelGameObject.AddComponent<Text>();
            snapshotIdLabelText.text = "Snapshot ID: " + _snapshotIdActiveInSettingsPanel;
            snapshotIdLabelText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            snapshotIdLabelText.color = Color.black;
            snapshotIdLabelText.fontSize = 20;
            snapshotIdLabelText.alignment = TextAnchor.UpperLeft;
            RectTransform snapshotIdLabelRect = snapshotIdLabelText.GetComponent<RectTransform>();
            snapshotIdLabelRect.sizeDelta = new Vector2(200, 30);
            snapshotIdLabelRect.anchorMin = new Vector2(0, 1);
            snapshotIdLabelRect.anchorMax = new Vector2(0, 1);
            snapshotIdLabelRect.pivot = new Vector2(0, 1);
            snapshotIdLabelRect.anchoredPosition = new Vector2(10, -60);

            // Snapshot name label
            GameObject snapshotNameLabelGameObject = new GameObject("SnapshotNameLabel");
            snapshotNameLabelGameObject.transform.SetParent(snapshotSettingsPanelGameObject.transform, false);
            Text snapshotNameLabelText = snapshotNameLabelGameObject.AddComponent<Text>();
            snapshotNameLabelText.text = "Snapshot Name:";
            snapshotNameLabelText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            snapshotNameLabelText.color = Color.black;
            snapshotNameLabelText.fontSize = 20;
            snapshotNameLabelText.alignment = TextAnchor.UpperLeft;
            RectTransform snapshotNameLabelRect = snapshotNameLabelText.GetComponent<RectTransform>();
            snapshotNameLabelRect.sizeDelta = new Vector2(200, 30);
            snapshotNameLabelRect.anchorMin = new Vector2(0, 1);
            snapshotNameLabelRect.anchorMax = new Vector2(0, 1);
            snapshotNameLabelRect.pivot = new Vector2(0, 1);
            snapshotNameLabelRect.anchoredPosition = new Vector2(10, -90); 

            // Snapshot name input field
            GameObject snapshotNameInputFieldGameObject = new GameObject("SnapshotNameInputField");
            snapshotNameInputFieldGameObject.transform.SetParent(snapshotSettingsPanelGameObject.transform, false);
            _snapshotNameInputField = snapshotNameInputFieldGameObject.AddComponent<InputField>();
            Image inputFieldImage = snapshotNameInputFieldGameObject.AddComponent<Image>();
            inputFieldImage.color = Color.white; // Set the background color of the input field
            RectTransform inputFieldRect = snapshotNameInputFieldGameObject.GetComponent<RectTransform>();
            inputFieldRect.sizeDelta = new Vector2(360, 30);
            inputFieldRect.anchorMin = new Vector2(0, 1);
            inputFieldRect.anchorMax = new Vector2(0, 1);
            inputFieldRect.pivot = new Vector2(0, 1);
            inputFieldRect.anchoredPosition = new Vector2(10, -120); 

            // Get Snapshot name
            string snapshotName = "";
            if (_snapshotIdActiveInSettingsPanel != -1) {
                var snapshot = _snapshotManager.GetSnapshot(_snapshotIdActiveInSettingsPanel);
                snapshotName = snapshot.Name;
            }

            // Setting up the text component of the input field
            GameObject inputFieldTextGameObject = new GameObject("Text");
            inputFieldTextGameObject.transform.SetParent(snapshotNameInputFieldGameObject.transform, false);
            Text inputFieldText = inputFieldTextGameObject.AddComponent<Text>();
            inputFieldText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            inputFieldText.text = ""; // Start with an empty text
            inputFieldText.color = Color.black;
            inputFieldText.alignment = TextAnchor.MiddleLeft;
            RectTransform inputFieldTextRect = inputFieldText.GetComponent<RectTransform>();
            inputFieldTextRect.anchorMin = new Vector2(0, 0);
            inputFieldTextRect.anchorMax = new Vector2(1, 1);
            inputFieldTextRect.sizeDelta = new Vector2(-20, 0); // Padding for the text inside the input field
            inputFieldTextRect.offsetMin = new Vector2(10, 6); // Left padding
            inputFieldTextRect.offsetMax = new Vector2(-10, -7); // Right padding

            // add snapshot name to input field
            _snapshotNameInputField.text = snapshotName;
        
            // Assign the text component to the input field
            _snapshotNameInputField.textComponent = inputFieldText;
            
            // Save button
            GameObject saveButtonGameObject = CreateButton(new Vector2(270, -160), "Save", _snapshotIdActiveInSettingsPanel, SolarizedColors.Green, null, OnSaveSnapshotSettingsClicked);
            saveButtonGameObject.transform.SetParent(snapshotSettingsPanelGameObject.transform, false);
            return snapshotSettingsPanelShadowGameObject;
        }

        public void UpdateUI(Dictionary<int, Snapshot> snapshots)
        {
            SetupUI();

            int i = 1;

            var sortedSnapshots = snapshots.OrderBy(entry => entry.Key);

            foreach (KeyValuePair<int, Snapshot> entry in sortedSnapshots)
            {
                var _settings = entry.Value.GetSettings();
                GameObject snapshotPanel = CreateSnapshotPanel(entry.Value.Name, entry.Key, new Vector2(10, -22 - (53 * i)), _settings);
                snapshotPanel.transform.SetParent(_panelGameObject.transform, false);
                i++;
            }
            // Settings panel
            if (_settingsPanelCreated) {
                 _settingsPanelGameObject.SetActive(false);
                GameObject.Destroy(_settingsPanelGameObject);
            }
            _settingsPanelGameObject = CreateSettingsPanel(new Vector2(830, -10));
            _settingsPanelGameObject.transform.SetParent(_rootGameObject.transform, false);
            _settingsPanelGameObject.SetActive(_settingsPanelActive);
            _settingsPanelCreated = true;
        }

        private void OnSettingsButtonClicked(int id)
        {
            Log.Info("Settings button clicked");
            _settingsPanelActive = !_settingsPanelActive;
            _settingsPanelGameObject.SetActive(_settingsPanelActive);
        }

        private void OnSnapshotSettingsButtonClicked(int id)
        {
            // toggle between -1 and id
            if (_snapshotIdActiveInSettingsPanel == -1) {
                _snapshotIdActiveInSettingsPanel = id;
                _snapshotSettingsPanelGameObject = CreateSnapshotSettingsPanel(new Vector2(830, -10));
                _snapshotSettingsPanelGameObject.transform.SetParent(_rootGameObject.transform, false);
                _snapshotSettingsPanelGameObject.SetActive(true);
            } else {
                _snapshotIdActiveInSettingsPanel = -1;
                GameObject.Destroy(_snapshotSettingsPanelGameObject);
            }
        }

        private void OnSaveSnapshotSettingsClicked(int id)
        {
            // Save snapshot name
            _snapshotManager.SaveSnapshotName(id, _snapshotNameInputField.text);
            OnSnapshotSettingsButtonClicked(id);
        }

        private void OnToggleClicked(string item, bool value)
        {
            Log.Info("Toggle clicked: " + item + " " + value);
            switch (item)
            {
                case "includeViewport":
                    _globalSettings.snapshotSettings.includePlayerView = value;
                    break;
                case "includePose":
                    _globalSettings.snapshotSettings.includePose = value;
                    break;
                case "includeMorphs":
                    _globalSettings.snapshotSettings.includeMorphs = value;
                    break;
                case "includeLights":
                    _globalSettings.snapshotSettings.includeLights = value;
                    break;
            }
        }   

        GameObject CreateSnapshotPanel(string snapshotName, int id, Vector2 pos, SnapshotSettings settings)
        {
            // Create a Panel
            GameObject panelGameObject = new GameObject("SnapshotPanel");
            panelGameObject.AddComponent<CanvasRenderer>();
            Image panelImage = panelGameObject.AddComponent<Image>();
            panelImage.color = SolarizedColors.LightTheme.Background; // Set panel color
            RectTransform panelRect = panelGameObject.GetComponent<RectTransform>();
            panelRect.sizeDelta = new Vector2(800-20, 50);
            panelRect.anchorMin = new Vector2(0, 1);
            panelRect.anchorMax = new Vector2(0, 1);
            panelRect.pivot = new Vector2(0, 1);
            panelRect.anchoredPosition = pos;

            // Icons
            if (settings.includePlayerView) {
                GameObject cameraIconGameObject = new GameObject("CameraIcon");
                cameraIconGameObject.transform.SetParent(panelGameObject.transform, false);
                Image cameraIconImage = cameraIconGameObject.AddComponent<Image>();
                cameraIconImage.sprite = Sprite.Create(_cameraIcon, new Rect(0, 0, _cameraIcon.width, _cameraIcon.height), new Vector2(0.5f, 0.5f));
                RectTransform cameraIconRect = cameraIconGameObject.GetComponent<RectTransform>();
                cameraIconRect.sizeDelta = new Vector2(25, 25);
                cameraIconRect.anchorMin = new Vector2(0, 1);
                cameraIconRect.anchorMax = new Vector2(0, 1);
                cameraIconRect.pivot = new Vector2(0, 1);
                cameraIconRect.anchoredPosition = new Vector2(10, -10);
            }

            if (settings.includePose) {
                GameObject poseIconGameObject = new GameObject("PoseIcon");
                poseIconGameObject.transform.SetParent(panelGameObject.transform, false);
                Image poseIconImage = poseIconGameObject.AddComponent<Image>();
                poseIconImage.sprite = Sprite.Create(_poseIcon, new Rect(0, 0, _poseIcon.width, _poseIcon.height), new Vector2(0.5f, 0.5f));
                RectTransform poseIconRect = poseIconGameObject.GetComponent<RectTransform>();
                poseIconRect.sizeDelta = new Vector2(25, 25);
                poseIconRect.anchorMin = new Vector2(0, 1);
                poseIconRect.anchorMax = new Vector2(0, 1);
                poseIconRect.pivot = new Vector2(0, 1);
                poseIconRect.anchoredPosition = new Vector2(40, -10);
            }

            if (settings.includeMorphs) {
                GameObject morphIconGameObject = new GameObject("MorphIcon");
                morphIconGameObject.transform.SetParent(panelGameObject.transform, false);
                Image morphIconImage = morphIconGameObject.AddComponent<Image>();
                morphIconImage.sprite = Sprite.Create(_morphIcon, new Rect(0, 0, _morphIcon.width, _morphIcon.height), new Vector2(0.5f, 0.5f));
                RectTransform morphIconRect = morphIconGameObject.GetComponent<RectTransform>();
                morphIconRect.sizeDelta = new Vector2(25, 25);
                morphIconRect.anchorMin = new Vector2(0, 1);
                morphIconRect.anchorMax = new Vector2(0, 1);
                morphIconRect.pivot = new Vector2(0, 1);
                morphIconRect.anchoredPosition = new Vector2(70, -10);
            }

            if (settings.includeLights) {
                GameObject lightsIconGameObject = new GameObject("LightsIcon");
                lightsIconGameObject.transform.SetParent(panelGameObject.transform, false);
                Image lightsIconImage = lightsIconGameObject.AddComponent<Image>();
                lightsIconImage.sprite = Sprite.Create(_lightBulbIcon, new Rect(0, 0, _lightBulbIcon.width, _lightBulbIcon.height), new Vector2(0.5f, 0.5f));
                RectTransform lightsIconRect = lightsIconGameObject.GetComponent<RectTransform>();
                lightsIconRect.sizeDelta = new Vector2(20, 25);
                lightsIconRect.anchorMin = new Vector2(0, 1);
                lightsIconRect.anchorMax = new Vector2(0, 1);
                lightsIconRect.pivot = new Vector2(0, 1);
                lightsIconRect.anchoredPosition = new Vector2(100, -10);
            }

            // Add snapshot name to panel
            GameObject snapshotNameGameObject = new GameObject("Text");
            snapshotNameGameObject.transform.SetParent(panelGameObject.transform, false);
            Text snapshotNameText = snapshotNameGameObject.AddComponent<Text>();
            snapshotNameText.text = snapshotName;
            snapshotNameText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            snapshotNameText.color = Color.black;
            snapshotNameText.fontSize = 18;
            snapshotNameText.alignment = TextAnchor.MiddleLeft;
            RectTransform snapshotNameRect = snapshotNameText.GetComponent<RectTransform>();
            snapshotNameRect.sizeDelta = new Vector2(280, 50);
            snapshotNameRect.anchorMin = new Vector2(0, 1);
            snapshotNameRect.anchorMax = new Vector2(0, 1);
            snapshotNameRect.pivot = new Vector2(0, 1);
            snapshotNameRect.anchoredPosition = new Vector2(130, 0);
            
            // Apply button
            int buttonStart = 735;
            int buttonMargin = 105;
            GameObject applyButtonGameObject = CreateButton(new Vector2(buttonStart-3*buttonMargin, -5), "Apply", id, SolarizedColors.Green, null, _onApplyClicked);
            applyButtonGameObject.transform.SetParent(panelGameObject.transform, false);

            // Overwrite button
            GameObject overwriteButtonGameObject = CreateButton(new Vector2(buttonStart-2*buttonMargin, -5), "Overwrite", id, SolarizedColors.Yellow, null, _onOverwriteClicked);
            overwriteButtonGameObject.transform.SetParent(panelGameObject.transform, false);

            // Delete button
            GameObject deleteButtonGameObject = CreateButton(new Vector2(buttonStart-buttonMargin, -5), "Delete", id, SolarizedColors.Red, null, _onDeleteClicked);
            deleteButtonGameObject.transform.SetParent(panelGameObject.transform, false);

            // Setting button
            GameObject settingsButtonGameObject = CreateButton(new Vector2(buttonStart, -5), "", id, SolarizedColors.Cyan, _settingsIcon, OnSnapshotSettingsButtonClicked, true);
            settingsButtonGameObject.transform.SetParent(panelGameObject.transform, false);


            return panelGameObject;
        }

        GameObject CreateButton(Vector2 pos, string text, int id, Color32 buttonColor, Texture2D icon = null, Action<int> onClick = null, bool iconOnly=false)
        {
            // Create Button GameObject
            GameObject buttonGameObject = new GameObject(text + "Button");
            buttonGameObject.AddComponent<CanvasRenderer>();

            Image buttonImage = buttonGameObject.AddComponent<Image>();
            buttonImage.color = buttonColor;

            // Set button RectTransform
            RectTransform buttonRect = buttonGameObject.GetComponent<RectTransform>();
            buttonRect.sizeDelta = icon != null ? new Vector2(140, 50) : new Vector2(100, 40);
            if (iconOnly) {
                buttonRect.sizeDelta = new Vector2(40, 40);
            }
            buttonRect.anchorMin = buttonRect.anchorMax = buttonRect.pivot = new Vector2(0, 1);
            buttonRect.anchoredPosition = pos;

            Button button = buttonGameObject.AddComponent<Button>();
            button.GetComponent<Button>().onClick.AddListener(() => onClick(id));

            // Set button hover color and other properties
            ColorBlock cb = button.colors;
            //cb.normalColor = buttonColor;
            cb.highlightedColor = new Color32((byte)(buttonColor.r - 20), (byte)(buttonColor.g - 20), (byte)(buttonColor.b - 20), 255);
            button.colors = cb;

            // Add icon to button if provided
            if (icon != null)
            {
                GameObject iconGameObject = new GameObject("Icon");
                iconGameObject.transform.SetParent(buttonGameObject.transform, false);
                Image iconImage = iconGameObject.AddComponent<Image>();
                iconImage.sprite = Sprite.Create(icon, new Rect(0, 0, icon.width, icon.height), new Vector2(0.5f, 0.5f));
                RectTransform iconRect = iconGameObject.GetComponent<RectTransform>();
                iconRect.sizeDelta = new Vector2(32, 32);
                iconRect.anchorMin = iconRect.anchorMax = iconRect.pivot = new Vector2(0, 1);
                iconRect.anchoredPosition = new Vector2(8, -9);
                if (iconOnly) {
                    iconRect.anchoredPosition = new Vector2(4, -4);
                }
            }

            // Add text to the button
            GameObject buttonTextGameObject = new GameObject("Text");
            buttonTextGameObject.transform.SetParent(buttonGameObject.transform, false);
            Text buttonText = buttonTextGameObject.AddComponent<Text>();
            buttonText.text = text;
            buttonText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            buttonText.color = Color.black;
            buttonText.fontSize = 18;
            buttonText.alignment = TextAnchor.MiddleCenter;
            RectTransform buttonTextRect = buttonText.GetComponent<RectTransform>();
            buttonTextRect.sizeDelta = new Vector2(120, 30);
            buttonTextRect.anchorMin = buttonTextRect.anchorMax = buttonTextRect.pivot = new Vector2(0.5f, 0.5f);
            buttonTextRect.anchoredPosition = icon != null ? new Vector2(20, -2) : Vector2.zero;

            return buttonGameObject;
        }

        GameObject CreateInputField()
        {
            // InputField GameObject
            GameObject inputFieldGameObject = new GameObject("InputField");
            inputFieldGameObject.AddComponent<CanvasRenderer>();
            Image inputFieldImage = inputFieldGameObject.AddComponent<Image>();
            inputFieldImage.color = Color.white; // Set input field background color
            InputField inputField = inputFieldGameObject.AddComponent<InputField>();

            // InputField Text
            GameObject inputFieldTextGameObject = new GameObject("Text");
            inputFieldTextGameObject.transform.SetParent(inputFieldGameObject.transform, false);
            Text inputFieldText = inputFieldTextGameObject.AddComponent<Text>();
            inputFieldText.text = "";
            inputFieldText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            inputFieldText.color = Color.black;
            inputField.textComponent = inputFieldText;
            RectTransform inputFieldTextRect = inputFieldText.GetComponent<RectTransform>();
            inputFieldTextRect.anchorMin = Vector2.zero;
            inputFieldTextRect.anchorMax = Vector2.one;
            inputFieldTextRect.sizeDelta = Vector2.zero;
            inputFieldTextRect.offsetMin = new Vector2(10, 6);
            inputFieldTextRect.offsetMax = new Vector2(-10, -7);

            // Placeholder Text
            GameObject placeholderGameObject = new GameObject("Placeholder");
            placeholderGameObject.transform.SetParent(inputFieldGameObject.transform, false);
            Text placeholderText = placeholderGameObject.AddComponent<Text>();
            placeholderText.text = "Enter text...";
            placeholderText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            placeholderText.color = Color.gray;
            inputField.placeholder = placeholderText;
            RectTransform placeholderTextRect = placeholderText.GetComponent<RectTransform>();
            placeholderTextRect.anchorMin = Vector2.zero;
            placeholderTextRect.anchorMax = Vector2.one;
            placeholderTextRect.sizeDelta = Vector2.zero;
            placeholderTextRect.offsetMin = new Vector2(10, 6);
            placeholderTextRect.offsetMax = new Vector2(-10, -7);

            return inputFieldGameObject;
        }

        GameObject CreateToggle(string name, string id, bool toggleChecked, Vector2 pos)
        {
            // Toggle GameObject
            GameObject toggleGameObject = new GameObject(name + "_Toggle");
            toggleGameObject.AddComponent<CanvasRenderer>();
            Toggle toggle = toggleGameObject.AddComponent<Toggle>();
            Image toggleImage = toggleGameObject.AddComponent<Image>();
            toggleImage.color = SolarizedColors.LightTheme.Foreground; // Set toggle background color
            RectTransform toggleRect = toggleGameObject.GetComponent<RectTransform>();
            toggleRect.sizeDelta = new Vector2(300, 30);
            toggleRect.anchorMin = new Vector2(0, 1);
            toggleRect.anchorMax = new Vector2(0, 1);
            toggleRect.pivot = new Vector2(0, 1);
            toggleRect.anchoredPosition = pos;
            
            // Toggle border
            GameObject toggleBorderGameObject = new GameObject(name + "_Border");
            toggleBorderGameObject.transform.SetParent(toggleGameObject.transform, false);
            Image toggleBorderImage = toggleBorderGameObject.AddComponent<Image>();
            toggleBorderImage.color = Color.black;
            toggle.graphic = toggleBorderImage;
            RectTransform toggleBorderRect = toggleBorderGameObject.GetComponent<RectTransform>();
            toggleBorderRect.sizeDelta = new Vector2(27, 27);
            toggleBorderRect.anchorMin = new Vector2(0, 1);
            toggleBorderRect.anchorMax = new Vector2(0, 1);
            toggleBorderRect.pivot = new Vector2(0, 1);
            toggleBorderRect.anchoredPosition = Vector2.zero;


            // Toggle background
            GameObject toggleBackgroundGameObject = new GameObject(name + "_Background");
            toggleBackgroundGameObject.transform.SetParent(toggleGameObject.transform, false);
            Image toggleBackgroundImage = toggleBackgroundGameObject.AddComponent<Image>();
            toggleBackgroundImage.color = Color.white;
            toggle.graphic = toggleBackgroundImage;
            RectTransform toggleBackgroundRect = toggleBackgroundGameObject.GetComponent<RectTransform>();
            toggleBackgroundRect.sizeDelta = new Vector2(25, 25);
            toggleBackgroundRect.anchorMin = new Vector2(0, 1);
            toggleBackgroundRect.anchorMax = new Vector2(0, 1);
            toggleBackgroundRect.pivot = new Vector2(0, 1);
            toggleBackgroundRect.anchoredPosition = new Vector2(1, -1);

            // Toggle Checkmark
            GameObject checkmarkGameObject = new GameObject(name + "_Checkmark");
            checkmarkGameObject.transform.SetParent(toggleBackgroundGameObject.transform, false);
            Image checkmarkImage = checkmarkGameObject.AddComponent<Image>();
            checkmarkImage.sprite = Sprite.Create(_toggleOn, new Rect(0, 0, _toggleOn.width, _toggleOn.height), new Vector2(0.5f, 0.5f));
            toggle.graphic = checkmarkImage;
            RectTransform checkmarkRect = checkmarkGameObject.GetComponent<RectTransform>();
            checkmarkRect.sizeDelta = new Vector2(25, 25);
            checkmarkRect.anchorMin = new Vector2(0, 1);
            checkmarkRect.anchorMax = new Vector2(0, 1);
            checkmarkRect.pivot = new Vector2(0, 1);
            checkmarkRect.anchoredPosition = Vector2.zero;
            toggle.onValueChanged.AddListener((bool value) => OnToggleClicked(id, value));

            if (toggleChecked)
            {
                toggle.isOn = true;
            } else
            {
                toggle.isOn = false;
            }

            // Toggle Label
            GameObject toggleLabelGameObject = new GameObject(name + "_Label");
            toggleLabelGameObject.transform.SetParent(toggleGameObject.transform, false);
            Text toggleLabel = toggleLabelGameObject.AddComponent<Text>();
            toggleLabel.text = name;
            toggleLabel.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            toggleLabel.fontSize = 18;
            toggleLabel.color = Color.black;
            RectTransform toggleLabelRect = toggleLabelGameObject.GetComponent<RectTransform>();
            toggleLabelRect.sizeDelta = new Vector2(250, 30);
            toggleLabelRect.anchorMin = new Vector2(0, 1);
            toggleLabelRect.anchorMax = new Vector2(0, 1);
            toggleLabelRect.pivot = new Vector2(0, 1);
            toggleLabelRect.anchoredPosition = new Vector2(34, -5);

            return toggleGameObject;
        }

        // on destroy
        public void OnDestroy()
        {
            if (_rootGameObject == null) return;
            // Rename the root game object to "flag" it is being destroyed by GC
            // So this game object can be ignored when checking if plugin is already running 
            //(during plugin reload game object might still be found even it is destroyed here)
            _rootGameObject.name = _rootGameObject.name+"_DOOMED";
            _rootGameObject.SetActive(false);
            GameObject.Destroy(_rootGameObject);
            _rootGameObject = null;
        }
    }
}